public class BinTreeH<D> {
   NodeH<D> root;

   /**
    * finds a node with a given key in this tree 
    * @param key
    * @return null, if the root is null; 
    *    a node with key, if this exists;
    *    a leaf below which the key has to be inserted, otherwise 
    */
   NodeH<D> find(int key) {
      NodeH<D> current = this.root;
      NodeH<D> previous = null;
      while (current != null) {
         previous = current;
         if (key < current.key) {
            current = current.left;
         } else if (key == current.key) {
            return current;
         } else {
            current = current.right;
         }
      }
      return previous; 
   }

   public void put(int key, D data) {
      NodeH<D> node = find(key);
      if (node == null) { // root is null
         this.root = new NodeH<D>(key,data,null,null,null,0);
      } else if (node.key == key) { // overwrite
         node.data = data;
      } else { // new node below node
         NodeH<D> newNode = new NodeH<D>(key,data,null,null,node,0);
         if (key < node.key) {
            node.left = newNode;
         } else {
            node.right = newNode;
         }
      }
   }

   /**
    * delivers the node which stores
    * the maximal key in the subtree with root "node"
    * @param node
    */
   static <D> NodeH<D> maximum(NodeH<D> node) {
      while (node.right != null) {
         node = node.right;
      }
      return node;
   }

   /**
    * removes a given non-null node from this tree
    */
   void removeNode(NodeH<D> node) {
      if (node.left == null && node.right == null) {
         if (node.parent == null) {
            this.root = null;
         } else {
            if (node.key < node.parent.key) {
               node.parent.left = null;
            } else {
               node.parent.right = null;
            }
         }
      } else { // non-leaf             
         if (node.left == null || node.right == null) { 
            NodeH<D> child = node.left == null ? node.right : node.left;
            if (node.parent == null) { // delete root
               this.root = child;
            } else if (node.key < node.parent.key) {
               node.parent.left = child;
            } else {
               node.parent.right = child;
            }
            child.parent = node.parent;
         } else { // two children
            NodeH<D> maxLeft = maximum(node.left);
            removeNode(maxLeft);
            node.data = maxLeft.data;
            node.key = maxLeft.key;
         }
      }      
   }

   public void remove(int key) {
      NodeH<D> node = find(key);
      if (node != null && node.key == key) {
         removeNode(node);
      }
   }
   
   public static <D> BinTreeH<D> create(int[] a) {
      BinTreeH<D> t = new BinTreeH<D>();
      for (int i : a) {
         t.put(i, null);
      }
      return t;
   }
   
   /**
    * @return -1 for the empty tree, and the height otherwise.
    */
   public int height() {
      throw new RuntimeException("Implement me");
   }
   
   public String toString() {
      throw new RuntimeException("Implement me");
   }
   
   public static void main(String[] args) {
      BinTreeH<String> t = BinTreeH.create(new int[] {10,4,8,15,7,-2,4,6,11,3,18,12,9,2,13});
      System.out.println(t.toString());
   }

}
